from abc import abstractmethod, abstractclassmethod
from typing import Mapping, Any, Set, Union, Optional
import functools
import logging

from datapipelines import NotFoundError, UnsupportedError
from merakicommons.ghost import Ghost
from merakicommons.container import SearchableList, SearchError

from ..configuration import settings
from ..data import Region, Platform
from ..dto.common import DtoObject

try:
    import ujson as json
except ImportError:
    import json


LOGGER = logging.getLogger("core")


def provide_default_region(method):
    @functools.wraps(method)
    def default_region_wrapper(self=None, *args, **kwargs):
        region = kwargs.pop("region", None)
        platform = kwargs.pop("platform", None)
        if region is None:
            if platform is None:
                region = settings.default_region
            else:
                if isinstance(platform, Platform):
                    region = platform.region
                else:
                    region = Platform(platform).region
        else:
            if not isinstance(region, Region):
                region = Region(region)
        kwargs["region"] = region.value
        if self:
                return method(self, *args, **kwargs)
        else:
            return method(*args, **kwargs)
    return default_region_wrapper


def get_latest_version(region: Union[Region, str], endpoint: Optional[str]):
    from .staticdata.realm import Realms
    if endpoint is not None:
        return Realms(region=region).latest_versions[endpoint]
    else:
        return Realms(region=region).version


class CoreData(object):
    def __init__(self, **kwargs):
        self._dto = {}
        self._update(kwargs)

    @classmethod
    def from_dto(cls, dto):
        self = cls()
        self._dto = dto
        return self

    @property
    @abstractclassmethod
    def _renamed(cls) -> Mapping[str, str]:
        """A mapping of the new method names to the old. If a name wasn't changed, it is not in this mapping."""
        pass

    def __contains__(self, item):
        return self._renamed.get(item, item) in self._dto

    def __iter__(self):
        for key in dir(self):
            if key in self:
                yield key

    def __str__(self):
        # TODO Update this to return the correct stuff when we update the data for this layer.
        return str(self._dto)

    def _update(self, data: Mapping[str, Any]) -> None:
        for key, value in data.items():
            self._set(key, value)

    def _set(self, key, value):
        key = self._renamed.get(key, key)
        self._dto[key] = value


class DataObjectList(list, CoreData):
    def __str__(self):
        return list.__str__(self)

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args)
        CoreData.__init__(self, **kwargs)

    def from_dto(cls, dto: Union[list, DtoObject]):
        self = CoreData.from_dto(dto)
        SearchableList.__init__(self, dto)


class CheckCache(type):
    @provide_default_region
    def __call__(cls, *args, **kwargs):
        cache = settings.pipeline._cache
        if cache is not None:
            # Try to find the obj in the cache
            from ..datastores.uniquekeys import construct_query
            query = construct_query(cls, **kwargs)
            if hasattr(cls, "version") and "version" not in query and not cls.__name__ == "Realms":
                query["version"] = get_latest_version(region=query["region"], endpoint=None)
            try:
                return cache.get(cls, query=query)
            except (NotFoundError, UnsupportedError):
                pass

        # If the obj was not found in the cache (or if there is no cache), create a new instance
        LOGGER.debug("Creating new {} from {}".format(cls.__name__, set(kwargs.keys())))
        return super().__call__(*args, **kwargs)


class CassiopeiaObject(object):
    _renamed = {}

    def __init__(self, **kwargs):
        # Note: Dto names are not allowed to be passed in.
        self._data = {type: type() for type in self._data_types}
        self(**kwargs)

    @classmethod
    def from_data(cls, data: CoreData):
        assert data is not None
        self = type.__call__(cls)  # Manually skip the CheckCache (well, all metaclasss' __call__s) for ghost objects if they are created via this constructor. Maybe CassiopeiaGhost should overload this method instead? I didn't because I'd have to copy-paste all the below code and I want it all in one spot for now.
        self._data = {_type: _type() for _type in self._data_types}
        if data.__class__ not in self._data_types:
            raise TypeError("Wrong data type '{}' passed to '{}.from_data'".format(data.__class__.__name__, self.__class__.__name__))
        self._data[data.__class__] = data
        # Without the below for-loops, if a Champion is instantiated with a StaticData data with `id`
        # in it, and then `.free_to_play` is called on that object, the query won't have an `id`.
        other_types = self._data_types - {data.__class__}
        for _type in other_types:
            for key in data:
                value = getattr(data, key)
                if key in dir(_type):
                    self._data[_type]._set(key, value)
        return self

    def __str__(self) -> str:
        # This is a bit strange because we'll print a list of dict-like objects rather than one joined dict, but we've decided it's appropritate.
        result = {}
        for type, data in self._data.items():
            result[str(type)] = str(data)
        return str(result).replace("\\'", "'")

    @property
    @abstractmethod
    def _data_types(self) -> Set[type]:
        """The `CoreData`_ types that belongs to this core type."""
        pass

    def __call__(self, **kwargs) -> "CassiopeiaObject":
        """Updates `self` with `kwargs` and returns `self`.

        Useful for updating default API parameters, for example:

            for champion in cass.get_champions():
                champion(champData={"tags"}).tags  # only pulls the tag data
        """
        # Update underlying data and deconstruct any Enums the user passed in.
        found = False
        for key, value in kwargs.items():
            for type in self._data_types:
                # Don't allow dto names to be passed in.
                if key in dir(type):
                    u = {key: value}
                    self._data[type]._update(u)
                    found = True
            if not found:
                LOGGER.warning("When initializing {}, key `{}` is not in type(s) {}. Not set.".format(self.__class__.__name__, key, self._data_types))
        return self


class CassiopeiaGhost(CassiopeiaObject, Ghost, metaclass=CheckCache):
    def __load__(self, load_group: CoreData = None) -> None:
        if load_group is None:  # Load all groups
            if self._Ghost__all_loaded:
                raise ValueError("object has already been loaded.")
            for group in self._Ghost__load_groups:
                if not self._Ghost__is_loaded(group):
                    self.__load__(group)
        else:  # Load the specific load group
            if self._Ghost__is_loaded(load_group):
                raise ValueError("object has already been loaded.")
            query = self.__get_query__()
            if hasattr(self.__class__, "version") and "version" not in query and not self.__class__.__name__ == "Realms":
                query["version"] = get_latest_version(region=query["region"], endpoint=None)
            data = settings.pipeline.get(type=self._load_types[load_group], query=query)
            self.__load_hook__(load_group, data)

    @property
    def _load_types(self):
        return {t: t for t in self._data_types}

    @abstractmethod
    def __get_query__(self):
        pass

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        if not isinstance(data, CoreData):
            raise TypeError("expected subclass of CoreData, got {cls}".format(cls=data.__class__))
        self._data[load_group] = data


class CassiopeiaGhostList(SearchableList, CassiopeiaGhost):
    def __hash__(self):
        return id(self)

    @property
    def _Ghost__load_groups(self) -> Set[type]:
        # Since @Ghost.property isn't set, the object doesn't have any load groups. Define them as equivalent to _data_types.
        return self._data_types

    def __init__(self, *args, **kwargs):
        SearchableList.__init__(self, *args)
        CassiopeiaGhost.__init__(self, **kwargs)
        self.__triggered_load = False

    @classmethod
    def from_data(cls, data: Union[list, CoreData]):
        raise NotImplemented
        #self = CassiopeiaGhost.from_data(data)
        #from ..transformers.staticdata import StaticDataTransformer
        #SearchableList.__init__(self, [StaticDataTransformer.champion_data_to_core(None, c) for c in data])
        #return self

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        super().__load_hook__(load_group=load_group, data=data)
        # Since @Ghost.property isn't set, the ghost's __property won't set itself as loaded because it never gets called.
        # Therefore we manually set it as loaded here.
        self._Ghost__set_loaded(load_group)

    def __iter__(self):
        if not self._Ghost__all_loaded and not self.__triggered_load:
            # TODO Discuss the below line with Rob.
            # This seems like a weird hack, but it also seems required. The issue is that within the Cache I call a .put
            #  on all the items in this object after putting this object itself in the cache.
            #  To do that, I iterate over this object, calling this __iter__ method, which triggers a __load__ (two
            #  lines down). That load triggers a transformer (Core to Dto; idk what that's called but it's irrelevant
            #  because there are situations where it should be called, e.g. with a database). So the load triggers a
            #  transformer, which has a for loop, which triggers this __iter__.
            #  The process is therefore:  __iter__ -> __load__ -> transformer -> __iter__ ...
            #  and we have a recursion depth error. To circumvent it, I added this __triggered_load boolean. It fixes
            #  that issue and everything works as-expected.
            self.__triggered_load = True
            self.__load__()
        return super().__iter__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except (IndexError, SearchError):
            if not self._Ghost__all_loaded:
                self.__triggered_load = True
                self.__load__()
            return super().__getitem__(item)



"""
This file defines constants which
are used in the various tests to
validate the API functions.
"""


# Summoner name to be used for Standard API calls, defaulting to NA
SUMMONER_NAME = "Kalturi"

# Unknown Summoner, used to validate that an error is thrown
UNKNOWN_SUMMONER_NAME = "abcaklsjdlakakdjlsakjdsdjlaksjdla"

# Champion name used for validating champion-related endpoints
CHAMP_NAME = "Karma"



import time
from abc import abstractmethod
from typing import MutableMapping, Any, Union, TypeVar, Iterable, Type, Dict
from collections import defaultdict

from datapipelines import DataSource, PipelineContext, NotFoundError
from merakicommons.ratelimits import RateLimiter, FixedWindowRateLimiter, MultiRateLimiter

from ..common import HTTPClient, HTTPError, Curl
from ...data import Platform


class APIRequestError(HTTPError):
    pass


class APIError(HTTPError):
    pass


class APINotFoundError(HTTPError):
    pass


class APIForbiddenError(APINotFoundError):
    pass


_ERROR_CODES = {
    400: APIRequestError,
    401: APIForbiddenError,
    403: APIRequestError,
    404: APINotFoundError,
    415: RuntimeError,
    429: RuntimeError,
    500: APIError,
    503: APIError
}

T = TypeVar("T")


_METHOD_RATE_LIMITS = defaultdict(lambda: (20000, 10))
_METHOD_RATE_LIMITS.update({
    "matches/id": (500, 10),
    "timelines/by-match/id": (500, 10),
    "matchlists/by-account/accountId": (1000, 10),
    "matchlists/by-account/accountId/recent": (1000, 10),
    "staticdata/champion": (10, 3600),
    "staticdata/champions": (10, 3600),
    "staticdata/versions": (10, 3600),
    "staticdata/mastery": (10, 3600),
    "staticdata/masteries": (10, 3600),
    "staticdata/rune": (10, 3600),
    "staticdata/runes": (10, 3600),
    "staticdata/item": (10, 3600),
    "staticdata/items": (10, 3600),
    "staticdata/maps": (10, 3600),
    "staticdata/summoner-spell": (10, 3600),
    "staticdata/summoner-spells": (10, 3600),
    "staticdata/realms": (10, 3600),
    "staticdata/language": (10, 3600),
    "staticdata/language-strings": (10, 3600),
    "staticdata/profile-icons": (10, 3600)
})


def _split_rate_limit_header(header):
    rates = []
    for pw in header.split(","):
        pw = pw.split(":")
        p, w = pw
        p, w = int(p), int(w)
        rates.append((p,w))
    return rates


class RiotAPIService(DataSource):
    def __init__(self, api_key: str, http_client: HTTPClient = None, application_rate_limiters: Dict[Platform, RateLimiter] = None) -> None:
        if http_client is None:
            self._client = HTTPClient()
        else:
            self._client = http_client

        self._headers = {
            "X-Riot-Token": api_key
        }

        # Both the application and method rate limiters will be in the same rate limiter
        self._rate_limiters = application_rate_limiters or {}

    def _get_rate_limiter(self, platform: Platform, endpoint: str):
        try:
            limiter = self._rate_limiters[(platform, endpoint)]
        except KeyError:
            # TODO: Move to settings and don't force dev limits.
            # TODO This is a circular import and instead this functionality should somehow be moved into settings.
            from cassiopeia.configuration import settings
            limits = settings.rate_limits
            limiter = MultiRateLimiter(*(
                [FixedWindowRateLimiter(window, permits) for permits, window in limits] +
                [FixedWindowRateLimiter(*reversed(_METHOD_RATE_LIMITS[endpoint]))]
            ))
            self._rate_limiters[(platform, endpoint)] = limiter
        return limiter

    def _get(self, url: str, parameters: MutableMapping[str, Any] = None, rate_limiter: MultiRateLimiter = None, connection: Curl = None, backoff: float = 1.) -> Union[dict, list, Any]:
        try:
            body, response_headers = self._client.get(url, parameters, self._headers, rate_limiter, connection)

            # If Riot changes the # of permits allowed in their response headers, change our rate limiters.
            # If any windows aren't of the correct length, they won't be updated / changed, so we throw an error if we
            #   receive a window for which the user (or the method rate limiter) doesn't have a set number of permits.
            # TODO This will override any user settings for this specific window.
            # We are currently ignoring the X-*-Rate-Limit-Count headers and assuming our rate limiter logic agrees.
            found = False
            if "X-App-Rate-Limit" in response_headers:
                for permits, window in _split_rate_limit_header(response_headers["X-App-Rate-Limit"]):
                    for _rate_limiter in rate_limiter:
                        if window == _rate_limiter._window_seconds:
                            if permits != _rate_limiter._window_permits:
                                _rate_limiter.set_permits(permits)
                            found = True
                            break
                    if found:
                        break
                else:
                    raise RuntimeError("Users must provide a rate limit for each window that Riot provides a rate limit for. The rate limits we received were: {}".format(
                        _split_rate_limit_header(response_headers["X-App-Rate-Limit"])
                    ))
            found = False
            if "X-Method-Rate-Limit" in response_headers:
                for permits, window in _split_rate_limit_header(response_headers["X-Method-Rate-Limit"]):
                    for _rate_limiter in rate_limiter:
                        if window == _rate_limiter._window_seconds:
                            if permits != _rate_limiter._window_permits:
                                _rate_limiter.set_permits(permits)
                            found = True
                            break
                    if found:
                        break
                else:
                    raise RuntimeError("Users must provide a rate limit for each window that Riot provides a rate limit for. The rate limits we received were: {}".format(
                        _split_rate_limit_header(response_headers["X-Method-Rate-Limit"])
                    ))

            return body
        except HTTPError as error:
            try:
                new_error_type = _ERROR_CODES[error.code]
            except KeyError as error2:
                raise RuntimeError("Encountered unexpected HTTP error code {code} with message \"{message}\". Report this to the Cassiopeia team.".format(code=error.code, message=str(error))) from error2

            # Try to properly handling the 429 and retry the call after the appropriate time limit.
            if error.code == 429:
                if "X-Rate-Limit-Type" not in error.response_headers or error.response_headers["X-Rate-Limit-Type"] == "service":
                    # Back off for one or more seconds
                    time.sleep(backoff)
                    if backoff < 3:  # Backoff at most 3 times, then quit
                        print("INFO: Unexpected service rate limit, backing off for {} seconds.".format(backoff))
                        return self._get(url, parameters, rate_limiter, connection, backoff + 1)  # Backoff for 1 more second each time. This isn't exponential backoff but it's probably fine.
                else:
                    if error.response_headers["X-Rate-Limit-Type"] == "application":
                        print("WARNING: Unexpected 429 due to application rate limit.")
                        if backoff < 3:  # Backoff isn't directly, here we're using it simply to prevent blacklisting if something in our code is wrong
                            rate_limiter.restrict_for(error.response_headers["Retry-After"])
                            return self._get(url, parameters, rate_limiter, connection, backoff + 1)  # Backoff for 1 more second each time. This isn't exponential backoff but it's probably fine.
                    elif error.response_headers["X-Rate-Limit-Type"] == "method":
                        print("WARNING: Unexpected 429 due to method rate limit.")
                        if backoff < 3:  # Backoff isn't directly, here we're using it simply to prevent blacklisting if something in our code is wrong
                            rate_limiter.restrict_for(error.response_headers["Retry-After"])
                            return self._get(url, parameters, rate_limiter, connection, backoff + 1)  # Backoff for 1 more second each time. This isn't exponential backoff but it's probably fine.
                    else:
                        raise RuntimeError("Unknown response header value for `X-Rate-Limit-Type`: {}".format(error.response_headers["X-Rate-Limit-Type"]))

            # The above error handling didn't work, so raise an appropriate error.
            if new_error_type is RuntimeError:
                new_error = RuntimeError("Encountered an HTTP error code {code} with message \"{message}\" which should have already been handled. Report this to the Cassiopeia team.".format(code=error.code, message=str(error)))
            elif new_error_type is APIError:
                new_error = APIError("The Riot API experienced an internal error on the request. You may want to retry the request after a short wait or continue without the result. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APINotFoundError:
                new_error = APINotFoundError("The Riot API returned a NOT FOUND error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIRequestError:
                new_error = APIRequestError("The Riot API returned an error on the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIForbiddenError:
                new_error = APIForbiddenError("The Riot API returned a FORBIDDEN error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            else:
                new_error = new_error_type(str(error))

            raise new_error from error

    @abstractmethod
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @abstractmethod
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass



from time import time
from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform, Region, Season, Queue, SEASON_IDS, QUEUE_IDS
from ...dto.match import MatchDto, MatchListDto, TimelineDto

T = TypeVar("T")


def _get_current_time(query: MutableMapping[str, Any], context: PipelineContext = None) -> int:
    return int(time()) * 1000


class MatchAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_match_query = Query. \
        has("gameId").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(MatchDto)
    def get_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI._validate_get_match_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/match/v3/matches/{id}".format(platform=query["platform"].value.lower(), id=query["gameId"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "matches/id"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["gameId"] = query["gameId"]
        data["region"] = query["platform"].region.value
        return MatchDto(data)

    _validate_get_many_match_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(MatchDto)
    def get_many_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MatchDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI._validate_get_many_match_query(query, context)

        def generator():
            for id in query["ids"]:
                url = "https://{platform}.api.riotgames.com/lol/match/v3/matches/{id}".format(platform=query["platform"].value.lower(), id=id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "matches/id"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["gameId"] = id
                data["region"] = query["platform"].region.value
                yield MatchDto(data)

        return generator()

    _validate_get_match_list_query = Query. \
        has("account.id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("recent").with_default(False).also. \
        can_have("startTime").with_default(0).also. \
        can_have("endTime").with_default(_get_current_time, supplies_type=int).also. \
        can_have("startIndex").as_(int).and_("endIndex").as_(int).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable)

    @get.register(MatchListDto)
    def get_match_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchListDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI._validate_get_match_list_query(query, context)

        if query["recent"]:
            url = "https://{platform}.api.riotgames.com/lol/match/v3/matchlists/by-account/{accountId}/recent".format(platform=query["platform"].value.lower(), accountId=query["account.id"])
            try:
                data = self._get(url, {})
            except APINotFoundError as error:
                raise NotFoundError(str(error)) from error

            data["accountId"] = query["account.id"]
            data["region"] = query["platform"].region.value
            return MatchListDto(data)
        else:
            params = {
                "beginTime": query["startTime"],
                "endTime": query["endTime"]
            }

            if "startIndex" in query and "endIndex" in query:
                params["beginIndex"] = query["startIndex"]
                params["endIndex"] = query["endIndex"]

            if "seasons" in query:
                seasons = {Season(season) for season in query["seasons"]}
                params["season"] = {SEASON_IDS[season] for season in seasons}
            else:
                seasons = set()

            if "champion.ids" in query:
                params["champion"] = {query["champion.ids"]}

            if "queues" in query:
                queues = {Queue(queue) for queue in query["queues"]}
                params["queue"] = {QUEUE_IDS[queue] for queue in queues}
            else:
                queues = set()

            url = "https://{platform}.api.riotgames.com/lol/match/v3/matchlists/by-account/{accountId}".format(platform=query["platform"].value.lower(), accountId=query["account.id"])
            try:
                data = self._get(url, params, self._get_rate_limiter(query["platform"], "matchlists/by-account/accountId"))
            except APINotFoundError as error:
                raise NotFoundError(str(error)) from error

            data["accountId"] = query["account.id"]
            data["region"] = query["platform"].region.value
            data["startTime"] = query["startTime"]
            data["endTime"] = query["endTime"]
            if "seasons" in query:
                data["seasons"] = seasons
            if "champion.ids" in query:
                data["champion"] = params["champion"]
            if "queues" in query:
                params["queue"] = queues
            for match in data["matches"]:
                match["accountId"] = query["account.id"]
                match["region"] = data["region"]
            return MatchListDto(data)

    _validate_get_many_match_list_query = Query. \
        has("account.ids").as_(Iterable).also. \
        has("platform").as_(Platform).also. \
        can_have("recent").with_default(False).also. \
        can_have("startTime").with_default(0).also. \
        can_have("endTime").with_default(_get_current_time, supplies_type=int).also. \
        can_have("startIndex").as_(int).and_("endIndex").as_(int).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable)

    @get_many.register(MatchListDto)
    def get_many_match_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[MatchListDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI._validate_get_many_match_list_query(query, context)

        if query["recent"]:
            def generator():
                for id in query["account.ids"]:
                    url = "https://{platform}.api.riotgames.com/lol/match/v3/matchlists/by-account/{accountId}/recent".format(platform=query["platform"].value.lower(), accountId=id)
                    try:
                        data = self._get(url, {}, self._get_rate_limiter(query["platform"], "matchlists/by-account/accountId/recent"))
                    except APINotFoundError as error:
                        raise NotFoundError(str(error)) from error

                    data["account.id"] = id
                    data["region"] = query["platform"].region.value
                    yield MatchListDto(data)
        else:
            params = {
                "beginTime": query["startTime"],
                "endTime": query["endTime"]
            }

            if "startIndex" in query and "endIndex" in query:
                params["beginIndex"] = query["startIndex"]
                params["endIndex"] = query["endIndex"]

            if "seasons" in query:
                seasons = {Season(season) for season in query["seasons"]}
                params["season"] = {SEASON_IDS[season] for season in seasons}
            else:
                seasons = set()

            if "champion.ids" in query:
                params["champion"] = {query["champion.ids"]}

            if "queues" in query:
                queues = {Queue(queue) for queue in query["queues"]}
                params["queue"] = {QUEUE_IDS[queue] for queue in queues}
            else:
                queues = set()

            def generator():
                for id in query["account.ids"]:
                    url = "https://{platform}.api.riotgames.com/lol/match/v3/matchlists/by-account/{accountId}".format(platform=query["platform"].value.lower(), accountId=id)
                    try:
                        data = self._get(url, params, self._get_rate_limiter(query["platform"], "matchlists/by-account/accountId"))
                    except APINotFoundError as error:
                        raise NotFoundError(str(error)) from error

                    data["account.id"] = id
                    data["region"] = query["platform"].region.value
                    data["startTime"] = query["startTime"]
                    data["endTime"] = query["endTime"]
                    if "seasons" in query:
                        data["seasons"] = seasons
                    if "champion.ids" in query:
                        data["champion"] = params["champion"]
                    if "queues" in query:
                        params["queue"] = queues
                    yield MatchListDto(data)

        return generator()

    _validate_get_timeline_query = Query. \
        has("matchId").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(TimelineDto)
    def get_match_timeline(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> TimelineDto:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI._validate_get_timeline_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/match/v3/timelines/by-match/{id}".format(platform=query["platform"].value.lower(), id=query["matchId"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "timelines/by-match/id"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["matchId"] = query["matchId"]
        data["region"] = query["platform"].region.value
        return TimelineDto(data)

    validate_get_many_timeline_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(TimelineDto)
    def get_many_match_timeline(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[TimelineDto, None, None]:
        if "region" in query and "platform" not in query:
            query["platform"] = Region(query["region"]).platform.value
        MatchAPI.validate_get_many_timeline_query(query, context)

        def generator():
            for id in query["ids"]:
                url = "https://{platform}.api.riotgames.com/lol/match/v3/timelines/by-match/{id}".format(platform=query["platform"].value.lower(), id=id)
                try:
                    data = self._get(url, {}, self._get_rate_limiter(query["platform"], "timelines/by-match/id"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["matchId"] = id
                data["region"] = query["platform"].region.value
                yield TimelineDto(data)

        return generator()



from typing import Tuple, Set, Union, MutableMapping, Any, Mapping, Iterable, Generator

from datapipelines import Query, PipelineContext, QueryValidationError

from ..data import Region, Platform, Queue, Tier

from ..dto.champion import ChampionListDto as ChampionStatusListDto, ChampionDto as ChampionStatusDto
from ..dto.championmastery import ChampionMasteryDto, ChampionMasteryListDto, ChampionMasteryScoreDto
from ..dto.league import LeagueListDto, LeaguePositionsDto
from ..dto.staticdata import ChampionDto, ChampionListDto, ItemDto, ItemListDto, LanguageStringsDto, LanguagesDto, MasteryDto, MasteryListDto, ProfileIconDataDto, ProfileIconDetailsDto, RealmDto, RuneDto, RuneListDto, SummonerSpellDto, SummonerSpellListDto, MapDto, MapListDto, VersionListDto
from ..dto.status import ShardStatusDto
from ..dto.masterypage import MasteryPagesDto, MasteryPageDto
from ..dto.match import MatchDto, MatchReferenceDto, TimelineDto
from ..dto.runepage import RunePagesDto, RunePageDto
from ..dto.spectator import CurrentGameInfoDto, FeaturedGamesDto
from ..dto.summoner import SummonerDto

from ..core.common import provide_default_region
from ..core.championmastery import ChampionMastery
from ..core.league import Leagues, ChallengerLeague, MasterLeague
from ..core.staticdata import Champion, Mastery, Rune, Item, SummonerSpell, Map, Locales, LanguageStrings, ProfileIcon, ProfileIcons, Realms, Versions, Items, Champions, Maps, SummonerSpells, Masteries, Runes
from ..core.status import ShardStatus
from ..core.masterypage import MasteryPage
from ..core.match import Match
from ..core.runepage import RunePage
from ..core.summoner import Summoner
from ..core.spectator import CurrentMatch, FeaturedMatches

from ..core.staticdata.champion import ChampionData
from ..core.summoner import Account, SummonerData

#############
# Utilities #
#############


def _rgetattr(obj, key):
    """Recursive getattr for handling dots in keys."""
    for k in key.split("."):
        obj = getattr(obj, k)
    return obj

def _hash_included_data(included_data: Set[str]) -> int:
    return hash(tuple(included_data))


def _get_default_version(query: Mapping[str, Any], context: PipelineContext) -> str:
    try:
        pipeline = context[PipelineContext.Keys.PIPELINE]
        versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
        return versions["versions"][0]
    except TypeError as error:
        raise KeyError("`version` must be provided in query") from error


def _get_default_locale(query: Mapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


def _region_to_platform_generator(regions: Iterable[Region]) -> Generator[Platform, None, None]:
    for region in regions:
        try:
            yield Region(region).platform
        except ValueError as e:
            raise QueryValidationError from e


def convert_region_to_platform(query: MutableMapping[str, Any]) -> None:
    if "region" in query and "platform" not in query:
        try:
            query["platform"] = Region(query["region"]).platform
        except ValueError as e:
            raise QueryValidationError from e

    if "regions" in query and "platforms" not in query:
        query["platforms"] = _region_to_platform_generator(query["regions"])


#########################################
# Construct Query From kwargs Functions #
#########################################


def construct_query(cls, **kwargs) -> dict:
    if cls is ChampionMastery:
        return construct_champion_mastery_query(**kwargs)
    if cls is Summoner:
        return construct_summoner_query(**kwargs)
    if cls is CurrentMatch:
        return construct_current_match_query(**kwargs)
    if cls is Leagues:
        return construct_leagues_query(**kwargs)
    else:
        return kwargs


@provide_default_region
def construct_summoner_query(*, id: int = None, account: Union[Account, int] = None, name: str = None, region: Union[Region, str]) -> dict:
    query = {"region": region}
    if id is not None:
        query["id"] = id
    if name is not None:
        query["name"] = name
    if account and isinstance(account, Account):
        query["account.id"] = account.id
    elif account is not None:
        query["account.id"] = account
    return query


@provide_default_region
def construct_champion_mastery_query(*, summoner: Union[Summoner, int, str], champion: Union[Champion, int, str], region: Union[Region, str]) -> dict:
    query = {"region": region}
    if isinstance(summoner, Summoner):
        summoner_data = summoner._data[SummonerData]
        try:
            query["summoner.id"] = summoner_data.id
        except KeyError:
            try:
                query["summoner.account.id"] = summoner_data.account_id
            except KeyError:
                query["summoner.name"] = summoner_data.name
    elif isinstance(summoner, str):
        query["summoner.name"] = summoner
    else:  # int
        query["summoner.id"] = summoner

    if isinstance(champion, Champion):
        champion_data = champion._data[ChampionData]
        try:
            query["champion.id"] = champion_data.id
        except KeyError:
            query["champion.name"] = champion_data.name
    elif isinstance(champion, str):
        query["champion.name"] = champion
    else:  # int
        query["champion.id"] = champion
    return query


@provide_default_region
def construct_current_match_query(*, summoner: Union[Summoner, int, str], region: Union[Region, str]) -> dict:
    query = {"region": region}
    if isinstance(summoner, Summoner):
        query["summoner.id"] = summoner.id
    elif isinstance(summoner, int):  # int
        query["summoner.id"] = summoner
    elif isinstance(summoner, str):
        query["summoner.id"] = Summoner(name=summoner).id
    assert "summoner.id" in query
    return query


@provide_default_region
def construct_leagues_query(*, summoner: Union[Summoner, int, str], region: Union[Region, str]) -> dict:
    query = {"region": region}
    if isinstance(summoner, Summoner):
        query["summoner.id"] = summoner.id
    elif isinstance(summoner, int):  # int
        query["summoner.id"] = summoner
    elif isinstance(summoner, str):
        query["summoner.id"] = Summoner(name=summoner).id
    assert "summoner.id" in query
    return query


#######
# DTO #
#######


################
# Champion API #
################


validate_champion_status_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int)


validate_many_champion_status_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable)


def for_champion_status_dto(champion_status: ChampionStatusDto) -> Tuple[str, int]:
    return champion_status["platform"], champion_status["id"]


def for_champion_status_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["id"]


def for_many_champion_status_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, id
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_status_list_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_champion_status_list_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_champion_status_list_dto(champion_status_list: ChampionStatusListDto) -> str:
    return champion_status_list["platform"]


def for_champion_status_list_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_champion_status_list_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


########################
# Champion Mastery API #
########################


validate_champion_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(int).also. \
    has("championId").as_(int)


validate_many_champion_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(int).also. \
    has("championIds").as_(Iterable)


def for_champion_mastery_dto(champion_mastery: ChampionMasteryDto) -> Tuple[str, int, int]:
    return champion_mastery["platform"], champion_mastery["playerId"], champion_mastery["championId"]


def for_champion_mastery_dto_query(query: Query) -> Tuple[str, int, int]:
    return query["platform"].value, query["playerId"], query["championId"]


def for_many_champion_mastery_dto_query(query: Query) -> Generator[Tuple[str, int, int], None, None]:
    for champion_id in query["championIds"]:
        try:
            champion_id = int(champion_id)
            yield query["platform"].value, query["playerId"], champion_id
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_mastery_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(int)


validate_many_champion_mastery_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerIds").as_(Iterable)


def for_champion_mastery_list_dto(champion_mastery_list: ChampionMasteryListDto) -> Tuple[str, int]:
    return champion_mastery_list["platform"], champion_mastery_list["playerId"]


def for_champion_mastery_list_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["playerId"]


def for_many_champion_mastery_list_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["playerIds"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_mastery_score_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(int)


validate_many_champion_mastery_score_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerIds").as_(Iterable)


def for_champion_mastery_score_dto(champion_mastery_score: ChampionMasteryScoreDto) -> Tuple[str, int]:
    return champion_mastery_score["platform"], champion_mastery_score["playerId"]


def for_champion_mastery_score_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["playerId"]


def for_many_champion_mastery_score_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["playerIds"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


##############
# League API #
##############


validate_league_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue).also. \
    has("tier").as_(Tier).also. \
    has("name").as_(str)


validate_many_league_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue).also. \
    has("tier").as_(Tier).also. \
    has("names").as_(Iterable)


def for_league_list_dto(league_list: LeagueListDto) -> Tuple[str, str, str, str]:
    return league_list["platform"], league_list["queue"], league_list["tier"], league_list["name"]


def for_league_list_dto_query(query: Query) -> Tuple[str, str, str, str]:
    return query["platform"].value, query["queue"].value, query["tier"].value, query["name"]


def for_many_league_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, str], None, None]:
    for name in query["names"]:
        try:
            name = str(name)
            yield query["platform"].value, query["queue"].value, query["tier"].value, name
        except ValueError as e:
            raise QueryValidationError from e


validate_league_positions_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int)


validate_many_league_positions_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerIds").as_(Iterable)


def for_league_positions_dto(league_positions: LeaguePositionsDto) -> Tuple[str, int]:
    return league_positions["platform"], league_positions["summonerId"]


def for_league_positions_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["summonerId"]


def for_many_league_positions_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["summonerIds"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


###################
# Static Data API #
###################


# Champion

validate_champion_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_champion_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_champion_dto(champion: ChampionDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return champion["platform"], champion["version"], champion["locale"], _hash_included_data(champion["includedData"]), champion[identifier]


def for_champion_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_champion_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_champion_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_champion_list_dto(champion_list: ChampionListDto) -> Tuple[str, str, str, int]:
    return champion_list["platform"], champion_list["version"], champion_list["locale"], _hash_included_data(champion_list["includedData"])


def for_champion_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_champion_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Item


validate_item_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_item_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("name").as_(Iterable)


def for_item_dto(item: ItemDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return item["platform"], item["version"], item["locale"], _hash_included_data(item["includedData"]), item[identifier]


def for_item_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_item_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_item_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_item_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_item_list_dto(item_list: ItemListDto) -> Tuple[str, str, str, int]:
    return item_list["platform"], item_list["version"], item_list["locale"], _hash_included_data(item_list["includedData"])


def for_item_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_item_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Language


validate_language_strings_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_language_strings_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_language_strings_dto(language_strings: LanguageStringsDto) -> Tuple[str, str, str]:
    return language_strings["platform"], language_strings["version"], language_strings["locale"]


def for_language_strings_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_language_strings_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


validate_languages_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_languages_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_languages_dto(languages: LanguagesDto) -> str:
    return languages["platform"]


def for_languages_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_languages_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


# Map

validate_map_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("mapId").as_(int).or_("mapName").as_(str)


validate_many_map_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("mapIds").as_(Iterable).or_("mapNames").as_(Iterable)


def for_map_dto(map: MapDto, identifier: str = "mapId") -> Tuple[str, str, str, Union[int, str]]:
    return map["platform"], map["version"], map["locale"], map[identifier]


def for_map_dto_query(query: Query) -> Tuple[str, str, str, Union[int, str]]:
    identifier = "mapId" if "mapId" in query else "mapName"
    return query["platform"].value, query["version"], query["locale"], query[identifier]


def for_many_map_dto_query(query: Query) -> Generator[Tuple[str, str, str, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["mapIds"], int) if "mapIds" in query else (query["mapNames"], str)
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_map_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_map_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_map_list_dto(map_list: MapListDto) -> Tuple[str, str, str]:
    return map_list["platform"], map_list["version"], map_list["locale"]


def for_map_list_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_map_list_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


# Mastery

validate_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_mastery_dto(mastery: MasteryDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return mastery["platform"], mastery["version"], mastery["locale"], _hash_included_data(mastery["includedData"]), mastery[identifier]


def for_mastery_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"], query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_mastery_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_mastery_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_mastery_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_mastery_list_dto(mastery_list: MasteryListDto) -> Tuple[str, str, str, int]:
    return mastery_list["platform"], mastery_list["version"], mastery_list["locale"], _hash_included_data(mastery_list["includedData"])


def for_mastery_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"], query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_mastery_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Profile Icon


validate_profile_icon_data_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_profile_icon_data_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_profile_icon_data_dto(profile_icon_data: ProfileIconDataDto) -> Tuple[str, str, str]:
    return profile_icon_data["platform"], profile_icon_data["version"], profile_icon_data["locale"]


def for_profile_icon_data_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_profile_icon_data_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


validate_profile_icon_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("id").as_(int)


validate_many_profile_icon_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable)


def for_profile_icon_dto(profile_icon: ProfileIconDetailsDto) -> Tuple[str, str, str, int]:
    return profile_icon["platform"], profile_icon["version"], profile_icon["locale"], profile_icon["id"]


def for_profile_icon_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], query["id"]


def for_many_profile_icon_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query["version"], query["locale"], id
        except ValueError as e:
            raise QueryValidationError from e


# Realm


validate_realm_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_realm_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_realm_dto(realm: RealmDto) -> str:
    return realm["platform"]


def for_realm_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_realm_dto_query(query: Query) -> Generator[str, None, None]:
    return query["platform"].value


# Rune


validate_rune_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_rune_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_rune_dto(rune: RuneDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return rune["platform"], rune["version"], rune["locale"], _hash_included_data(rune["includedData"]), rune[identifier]


def for_rune_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_rune_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_rune_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_rune_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_rune_list_dto(rune_list: RuneListDto) -> Tuple[str, str, str, int]:
    return rune_list["platform"], rune_list["version"], rune_list["locale"], _hash_included_data(rune_list["includedData"])


def for_rune_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_rune_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Summoner Spell


validate_summoner_spell_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_summoner_spell_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_spell_dto(summoner_spell: SummonerSpellDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return summoner_spell["platform"], summoner_spell["version"], summoner_spell["locale"], _hash_included_data(summoner_spell["includedData"]), summoner_spell[identifier]


def for_summoner_spell_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_summoner_spell_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_summoner_spell_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_summoner_spell_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_summoner_spell_list_dto(summoner_spell_list: SummonerSpellListDto) -> Tuple[str, str, str, int]:
    return summoner_spell_list["platform"], summoner_spell_list["version"], summoner_spell_list["locale"], _hash_included_data(summoner_spell_list["includedData"])


def for_summoner_spell_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_summoner_spell_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Version


validate_version_list_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_version_list_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_version_list_dto(version_list: VersionListDto) -> str:
    return version_list["platform"]


def for_version_list_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_version_list_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


##############
# Status API #
##############


validate_shard_status_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_shard_status_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_shard_status_dto(shard_status: ShardStatusDto) -> str:
    return shard_status["platform"]


def for_shard_status_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_shard_status_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


#################
# Masteries API #
#################


validate_mastery_page_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").also. \
    has("id").as_(int)


validate_many_mastery_page_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").also. \
    has("ids").as_(Iterable)


def for_mastery_page_dto(mastery_page: MasteryPageDto) -> Tuple[str, int, int]:
    return mastery_page["platform"], mastery_page["summonerId"], mastery_page["id"]


def for_mastery_page_dto_query(query: Query) -> Tuple[str, int, int]:
    return query["platform"].value, query["summonerId"], query["id"]


def for_many_mastery_page_dto_query(query: Query) -> Generator[Tuple[str, int, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query["summonerId"], id
        except ValueError as e:
            raise QueryValidationError from e


validate_mastery_pages_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int)


validate_many_mastery_pages_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerIds").as_(Iterable)


def for_mastery_pages_dto(mastery_pages: MasteryPagesDto) -> Tuple[str, int]:
    return mastery_pages["platform"], mastery_pages["summonerId"]


def for_mastery_pages_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["summonerId"]


def for_many_mastery_pages_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["summonerIds"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


#############
# Match API #
#############


validate_match_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_match_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_match_dto(match: MatchDto) -> Tuple[str, int]:
    return match["platform"], match["gameId"]


def for_match_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_match_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_match_reference_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_match_reference_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_match_reference_dto(match_reference: MatchReferenceDto) -> Tuple[str, int]:
    return match_reference["platform"], match_reference["gameId"]


def for_match_reference_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_match_reference_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_match_timeline_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("matchId").as_(int)


validate_many_match_timeline_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("matchIds").as_(Iterable)


def for_match_timeline_dto(match_timeline: TimelineDto) -> Tuple[str, int]:
    return match_timeline["platform"], match_timeline["matchId"]


def for_match_timeline_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["matchId"]


def for_many_match_timeline_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for match_id in query["matchIds"]:
        try:
            match_id = int(match_id)
            yield query["platform"].value, match_id
        except ValueError as e:
            raise QueryValidationError from e


#############
# Runes API #
#############


validate_rune_page_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int).also. \
    has("id").as_(int)


validate_many_rune_page_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").also. \
    has("ids").as_(Iterable)


def for_rune_page_dto(rune_page: RunePageDto) -> Tuple[str, int, int]:
    return rune_page["platform"], rune_page["summonerId"], rune_page["id"]


def for_rune_page_dto_query(query: Query) -> Tuple[str, int, int]:
    return query["platform"].value, query["summonerId"], query["id"]


def for_many_rune_page_dto_query(query: Query) -> Generator[Tuple[str, int, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query["summonerId"], id
        except ValueError as e:
            raise QueryValidationError from e


validate_rune_pages_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int)


validate_many_rune_pages_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerIds").as_(Iterable)


def for_rune_pages_dto(rune_pages: RunePagesDto) -> Tuple[str, int]:
    return rune_pages["platform"], rune_pages["summonerId"]


def for_rune_pages_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["summonerId"]


def for_many_rune_pages_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["summonerIds"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


#################
# Spectator API #
#################


validate_current_game_info_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_current_game_info_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_current_game_info_dto(current_game_info: CurrentGameInfoDto) -> Tuple[str, int]:
    return current_game_info["platform"], current_game_info["gameId"]


def for_current_game_info_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_current_game_info_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_featured_game_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_featured_game_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_featured_games_dto(featured_games: FeaturedGamesDto) -> str:
    return featured_games["platform"]


def for_featured_games_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_featured_games_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


################
# Summoner API #
################


validate_summoner_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int).or_("accountId").as_(int).or_("name").as_(str)


validate_many_summoner_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable).or_("accountIds").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_dto(summoner: SummonerDto, identifier: str = "id") -> Tuple[str, Union[int, str]]:
    return summoner["platform"], summoner[identifier]


def for_summoner_dto_query(query: Query) -> Tuple[str, Union[int, str]]:
    if "id" in query:
        identifier = "id"
    elif "accountId" in query:
        identifier = "accountId"
    else:
        identifier = "name"
    return query["platform"].value, query[identifier]


def for_many_summoner_dto_query(query: Query) -> Generator[Tuple[str, Union[int, str]], None, None]:
    if "ids" in query:
        identifiers, identifier_type = query["ids"], int
    elif "accountIds" in query:
        identifiers, identifier_type = query["accountIds"], int
    else:
        identifiers, identifier_type = query["names"], str
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, identifier
        except ValueError as e:
            raise QueryValidationError from e


########
# Core #
########


##########################################################
# Champion API (Covered by Static Data Champion in core) #
##########################################################


########################
# Champion Mastery API #
########################


validate_champion_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int).or_("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("championId").as_(int).or_("champion.id").as_(int).or_("champion.name").as_(str)


validate_many_champion_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("champions.id").as_(Iterable).or_("champions.name").as_(Iterable)


def for_champion_mastery(champion_mastery: ChampionMastery, summoner_identifier: str = "id", champion_identifier: str = "id") -> Tuple[str, Union[int, str], Union[int, str]]:
    return champion_mastery.platform.value, _rgetattr(champion_mastery.summoner, summoner_identifier), _rgetattr(champion_mastery.champion, champion_identifier)


def for_champion_mastery_query(query: Query) -> Tuple[str, Union[int, str], Union[int, str]]:
    if "playerId" in query:
        summoner_identifier = "playerId"
    elif "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"

    if "championId" in query:
        champion_identifier = "championId"
    elif "champion.id" in query:
        champion_identifier = "champion.id"
    else:
        champion_identifier = "champion.name"

    return query["platform"].value, query[summoner_identifier], query[champion_identifier]


def for_many_champion_mastery_query(query: Query) -> Generator[Tuple[str, Union[int, str], Union[int, str]], None, None]:
    if "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"

    champion_identifiers, champion_identifier_type = (query["champions.id"], int) if "champions.id" in query else (query["champions.name"], str)

    for champion_identifier in champion_identifiers:
        try:
            champion_identifier = champion_identifier_type(champion_identifier)
            yield query["platform"].value, query[summoner_identifier], champion_identifier
        except ValueError as e:
            raise QueryValidationError from e


##############
# League API #
##############


validate_leagues_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(int)


validate_many_leagues_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoners.id").as_(Iterable)


def for_leagues(leagues: Leagues, summoner_identifier: str = "summoner.id") -> Tuple[str, int]:
    return leagues.platform.value, leagues._Leagues__summoner.id


def for_leagues_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["summoner.id"]


def for_many_leagues_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for id in query["summoners.id"]:
        try:
            yield query["platform"].value, id
        except ValueError as e:
            raise QueryValidationError from e

# Challenger

validate_challenger_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue)


validate_many_challenger_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queues").as_(Iterable)


def for_challenger_league(league: ChallengerLeague, queue_identifier: str = "value") -> Tuple[str, str]:
    return league.platform.value, league.queue.value


def for_challenger_league_query(query: Query) -> Tuple[str, str]:
    return query["platform"].value, query["queue"].value


def for_many_challenger_league_query(query: Query) -> Generator[Tuple[str, str], None, None]:
    for queue in query["queues"]:
        try:
            yield query["platform"].value, queue.value
        except ValueError as e:
            raise QueryValidationError from e

# Master

validate_master_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue)


validate_many_master_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queues").as_(Iterable)


def for_master_league(league: MasterLeague, queue_identifier: str = "value") -> Tuple[str, str]:
    return league.platform.value, league.queue.value


def for_master_league_query(query: Query) -> Tuple[str, str]:
    return query["platform"].value, query["queue"].value


def for_many_master_league_query(query: Query) -> Generator[Tuple[str, str], None, None]:
    for queue in query["queues"]:
        try:
            yield query["platform"].value, queue.value
        except ValueError as e:
            raise QueryValidationError from e


###################
# Static Data API #
###################


# Champion

validate_champion_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_champion_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_champion(champion: Champion, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return champion.platform.value, champion.version, champion.locale, _hash_included_data(champion.included_data), _rgetattr(champion, identifier)


def for_champion_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash, query[identifier]


def for_many_champion_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["included_data"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_champions_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"})


validate_many_champions_query = Query. \
    has("platforms").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"})


def for_champions(champions: Champions, identifier: str = None) -> Tuple[str, str, str, int]:
    return champions.platform.value, champions.version, champions.locale, _hash_included_data(champions.included_data)


def for_champions_query(query: Query) -> Tuple[str, str, str, int]:
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash


def for_many_champions_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["included_data"])
    for platform in query["platforms"]:
        try:
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Item

validate_item_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_item_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_item(item: Item, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return item.platform.value, item.version, item.locale, _hash_included_data(item.included_data), _rgetattr(item, identifier)


def for_item_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash, query[identifier]


def for_many_item_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["included_data"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e

validate_items_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"})


validate_many_items_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"})


def for_items(items: Items, identifier: str = "platform") -> Tuple[str, str, str, int]:
    return items.platform.value, items.version, items.locale, _hash_included_data(items.included_data)


def for_items_query(query: Query) -> Tuple[str, str, str, int]:
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash


def for_many_items_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["included_data"])
    for platform in query["platforms"]:
        try:
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Language

validate_languages_query = Query. \
    has("platform").as_(Platform)


validate_many_languages_query = Query. \
    has("platforms").as_(Iterable)


def for_languages(languages: Locales) -> str:
    return languages.platform.value


def for_languages_query(query: Query) -> str:
    return query["platform"].value


def for_many_languages_query(query: Query) -> Generator[str, None, None]:
    return query["platform"].value


validate_language_strings_query = Query. \
    has("platform").as_(Platform)


validate_many_language_strings_query = Query. \
    has("platforms").as_(Iterable)


def for_language_strings(languages: LanguageStrings) -> str:
    return languages.platform.value


def for_language_strings_query(query: Query) -> str:
    return query["platform"].value


def for_many_language_strings_query(query: Query) -> Generator[str, None, None]:
    return query["platform"].value


# Map

validate_map_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("mapId").as_(int).or_("id").as_(int).or_("name").as_(str)


validate_many_map_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_map(map: Map, identifier: str = "id") -> Tuple[str, str, str, Union[int, str]]:
    return map.platform.value, map.version, map.locale, _rgetattr(map, identifier)


def for_map_query(query: Query) -> Tuple[str, str, str, Union[int, str]]:
    if "mapId" in query:
        identifier = "mapId"
    elif "id" in query:
        identifier = "id"
    else:
        identifier = "name"
    return query["platform"].value, query["version"], query["locale"], query[identifier]


def for_many_map_query(query: Query) -> Generator[Tuple[str, str, str, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_maps_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_maps_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_maps(maps: Maps, identifier: str = "id") -> Tuple[str, str, str]:
    return maps.platform.value, maps.version, maps.locale


def for_maps_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_maps_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        yield platform.value, query["version"], query["locale"]


# Mastery

validate_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_mastery(mastery: Mastery, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return mastery.platform.value, mastery.version, mastery.locale, _hash_included_data(mastery.included_data), _rgetattr(mastery, identifier)


def for_mastery_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash, query[identifier]


def for_many_mastery_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["included_data"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_masteries_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"})


validate_many_masteries_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"})


def for_masteries(masteries: Masteries, identifier: str = "platform") -> Tuple[str, str, str, int]:
    return masteries.platform.value, masteries.version, masteries.locale, _hash_included_data(masteries.included_data)


def for_masteries_query(query: Query) -> Tuple[str, str, str, int]:
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash


def for_many_masteries_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["included_data"])
    for platform in query["platforms"]:
        try:
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Profile Icon

validate_profile_icons_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_profile_icons_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_profile_icons(profile_icon: ProfileIcons) -> Tuple[str, str, str]:
    return profile_icon["platform"], profile_icon["version"], profile_icon["locale"]


def for_profile_icons_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_profile_icons_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


validate_profile_icon_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("id").as_(int)


validate_many_profile_icon_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable)


def for_profile_icon(profile_icon: ProfileIcon) -> Tuple[str, str, str, int]:
    return profile_icon["platform"], profile_icon["version"], profile_icon["locale"], profile_icon["id"]


def for_profile_icon_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], query["id"]


def for_many_profile_icon_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query["version"], query["locale"], id
        except ValueError as e:
            raise QueryValidationError from e


# Realm

validate_realms_query = Query. \
    has("platform").as_(Platform)


validate_many_realms_query = Query. \
    has("platforms").as_(Iterable)


def for_realms(realm: Realms) -> str:
    return realm.platform.value


def for_realms_query(query: Query) -> str:
    return query["platform"].value


def for_many_realms_query(query: Query) -> Generator[str, None, None]:
    return query["platform"].value


# Rune

validate_rune_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_rune_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_rune(rune: Rune, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return rune.platform.value, rune.version, rune.locale, _hash_included_data(rune.included_data), _rgetattr(rune, identifier)


def for_rune_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash, query[identifier]


def for_many_rune_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["included_data"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_runes_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"})


validate_many_runes_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"})


def for_runes(runes: Runes, identifier: str = "platform") -> Tuple[str, str, str, int]:
    return runes.platform.value, runes.version, runes.locale, _hash_included_data(runes.included_data)


def for_runes_query(query: Query) -> Tuple[str, str, str, int]:
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash


def for_many_runes_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["included_data"])
    for platform in query["platforms"]:
        try:
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Summoner Spell

validate_summoner_spell_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_summoner_spell_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_spell(summoner_spell: SummonerSpell, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return summoner_spell.platform.value, summoner_spell.version, summoner_spell.locale, _hash_included_data(summoner_spell.included_data), _rgetattr(summoner_spell, identifier)


def for_summoner_spell_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash, query[identifier]


def for_many_summoner_spell_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["included_data"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_summoner_spells_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").as_(set).or_("included_data").with_default({"all"})


validate_many_summoner_spells_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("included_data").with_default({"all"})


def for_summoner_spells(summoner_spells: SummonerSpells, identifier: str = "platform") -> Tuple[str, str, str, int]:
    return summoner_spells.platform.value, summoner_spells.version, summoner_spells.locale, _hash_included_data(summoner_spells.included_data)


def for_summoner_spells_query(query: Query) -> Tuple[str, str, str, int]:
    included_data_hash = _hash_included_data(query["includedData"]) if "includedData" in query else _hash_included_data(query["included_data"])
    return query["platform"].value, query["version"], query["locale"], included_data_hash


def for_many_summoner_spells_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["included_data"])
    for platform in query["platforms"]:
        try:
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Version

validate_versions_query = Query. \
    has("platform").as_(Platform)


validate_many_versions_query = Query. \
    has("platforms").as_(Iterable)


def for_versions(versions: Versions) -> str:
    return versions.platform.value


def for_versions_query(query: Query) -> str:
    return query["platform"].value


def for_many_versions_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


##############
# Status API #
##############

validate_shard_status_query = Query. \
    has("platform").as_(Platform)


validate_many_shard_status_query = Query. \
    has("platforms").as_(Iterable)


def for_shard_status(shard_status: ShardStatus) -> str:
    return shard_status.platform.value


def for_shard_status_query(query: Query) -> str:
    return query["platform"].value


def for_many_shard_status_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


#################
# Masteries API #
#################


validate_mastery_page_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int).or_("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("id").as_(int)


validate_many_mastery_page_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("ids").as_(Iterable)


def for_mastery_page(mastery_page: MasteryPage, summoner_identifier: str = "id") -> Tuple[str, Union[int, str], int]:
    return mastery_page.platform.value, _rgetattr(mastery_page.summoner, summoner_identifier), mastery_page.id


def for_mastery_page_query(query: Query) -> Tuple[str, Union[int, str], int]:
    if "summonerId" in query:
        summoner_identifier = "summonerId"
    elif "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"
    return query["platform"].value, query[summoner_identifier], query["id"]


def for_many_mastery_page_query(query: Query) -> Generator[Tuple[str, Union[int, str], int], None, None]:
    if "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"

    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query[summoner_identifier], id
        except ValueError as e:
            raise QueryValidationError from e


#############
# Match API #
#############


validate_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int).or_("id").as_(int)


validate_many_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable)


def for_match(match: Match) -> Tuple[str, int]:
    return match.platform.value, match.id


def for_match_query(query: Query) -> Tuple[str, int]:
    identifier = "gameId" if "gameId" in query else "id"
    return query["platform"].value, query[identifier]


def for_many_match_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, id
        except ValueError as e:
            raise QueryValidationError from e


#############
# Runes API #
#############


validate_rune_page_query = Query. \
    has("platform").as_(Platform).also. \
    has("summonerId").as_(int).or_("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("id").as_(int)


validate_many_rune_page_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(int).or_("summoner.account.id").as_(int).or_("summoner.name").as_(str).also. \
    has("ids").as_(Iterable)


def for_rune_page(rune_page: RunePage, summoner_identifier: str = "id") -> Tuple[str, int, int]:
    return rune_page.platform.value, _rgetattr(rune_page.summoner, summoner_identifier), rune_page.id


def for_rune_page_query(query: Query) -> Tuple[str, Union[int, str], int]:
    if "summonerId" in query:
        summoner_identifier = "summonerId"
    elif "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"
    return query["platform"].value, query[summoner_identifier], query["id"]


def for_many_rune_page_query(query: Query) -> Generator[Tuple[str, Union[int, str], int], None, None]:
    if "summoner.id" in query:
        summoner_identifier = "summoner.id"
    elif "summoner.account.id" in query:
        summoner_identifier = "summoner.account.id"
    else:
        summoner_identifier = "summoner.name"

    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query[summoner_identifier], id
        except ValueError as e:
            raise QueryValidationError from e


#################
# Spectator API #
#################

validate_current_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(int)


validate_many_current_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.ids").as_(Iterable)


def for_current_match(current_match_info: CurrentMatch) -> Tuple[str, int]:
    return current_match_info.platform, current_match_info._CurrentMatch__summoner.id


def for_current_match_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["summoner.id"]


def for_many_current_match_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for summoner_id in query["summoner.ids"]:
        try:
            summoner_id = int(summoner_id)
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


validate_featured_matches_query = Query. \
    has("platform").as_(Platform)


validate_many_featured_matches_query = Query. \
    has("platforms").as_(Iterable)


def for_featured_matches(featured_matches: FeaturedMatches) -> str:
    return featured_matches.platform


def for_featured_matches_query(query: Query) -> str:
    return query["platform"].value


def for_many_featured_matches_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


################
# Summoner API #
################


validate_summoner_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int).or_("account.id").as_(int).or_("name").as_(str).or_("accountId").as_(int)


validate_many_summoner_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable).or_("account.id").as_(Iterable).or_("names").as_(Iterable)


def for_summoner(summoner: Summoner, identifier: str = "id") -> Tuple[str, Union[int, str]]:
    return summoner.platform.value, _rgetattr(summoner, identifier)


def for_summoner_query(query: Query) -> Tuple[str, Union[int, str]]:
    if "id" in query:
        identifier = "id"
    elif "account.id" in query:
        identifier = "account.id"
    elif "name" in query:
        identifier = "name"
    else:
        identifier = "accountId"
    return query["platform"].value, query[identifier]


def for_many_summoner_query(query: Query) -> Generator[Tuple[str, Union[int, str]], None, None]:
    if "ids" in query:
        identifiers, identifier_type = query["ids"], int
    elif "account.id" in query:
        identifiers, identifier_type = query["account.id"], int
    else:
        identifiers, identifier_type = query["names"], str

    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, identifier
        except ValueError as e:
            raise QueryValidationError from e



import sys
import os

try:
    import ujson as json
except ImportError:
    import json


head, tail = os.path.split(__file__)
if len(sys.argv) > 1 and any(fn.endswith(".json") for fn in sys.argv):
    for fn in sys.argv:
        if fn.endswith(".json"):
            filename = fn
            break
else:
    filename = os.path.join(head, "default.json")

if not os.path.exists(filename):
    filename = os.path.join(head, filename)
    if not os.path.exists(filename):
        filename = os.path.join(head, "default.json")

config = json.loads(open(filename).read())
if "logging" not in config:
    config["logging"] = {}
assert "Riot API" in config and "key" in config["Riot API"]


from .cassiopeia import get_realms, get_challenger_league, get_champion_masteries, get_champion, get_champion_mastery, get_champions, get_current_match, get_featured_matches, get_items, get_language_strings, get_locales, get_leagues, get_maps, get_master_league, get_masteries, get_mastery_pages, get_match, get_match_history, get_patch, get_profile_icons, get_rune_pages, get_runes, get_status, get_summoner, get_summoner_spells, get_version, get_versions
from .core import Champion, Champions, Rune, Runes, Mastery, Masteries, Item, Items, SummonerSpell, SummonerSpells, ProfileIcon, ProfileIcons, Versions, Maps, Summoner, Account, ChampionMastery, ChampionMasteries, Match, FeaturedMatches, ShardStatus, ChallengerLeague, MasterLeague, Map, Realms, LanguageStrings, Locales



import cassiopeia
import pytest

from merakicommons.container import SearchableList
from datapipelines.common import NotFoundError

from .constants import CHAMP_NAME, SUMMONER_NAME, UNKNOWN_SUMMONER_NAME


def test_masteries_correct_type():
    # TODO: Once get_champion_masteries works with passing a name,
    #       replace this with get_champion_masteries(SUMMONER_NAME)
    summoner = cassiopeia.get_summoner(name=SUMMONER_NAME)
    champ_masteries = cassiopeia.get_champion_masteries(summoner.id)

    assert isinstance(champ_masteries, SearchableList)
    assert all(isinstance(cm, cassiopeia.ChampionMastery) for cm in champ_masteries)


def test_returns_empty_list_with_unknown_summoner():
    # TODO: Also do this with name, once it's implemented (see above)
    cm = cassiopeia.get_champion_masteries(123901923180391)
    assert cm == []


# def test_masteries_with_id_or_summoner():
    # TODO
    # Comment out these tests once get_champion_masteries supports
    # retrieving mastery through name or summoner again
    # (currently only supports it through ID)

    # summ = cassiopeia.get_summoner(name=SUMMONER_NAME)

    # from_summoner = set(cassiopeia.get_champion_masteries(summ))
    # from_id = set(cassiopeia.get_champion_masteries(summ.id))
    # from_name = set(cassiopeia.get_champion_masteries(summ.name))

    # assert from_summoner == from_id
    # assert from_id == from_name


def test_mastery_return():
    summ = cassiopeia.get_summoner(name=SUMMONER_NAME)
    champ = cassiopeia.get_champion(CHAMP_NAME)
    champ_mastery = cassiopeia.get_champion_mastery(summ.id, champion=champ)

    assert isinstance(champ_mastery, cassiopeia.ChampionMastery)
    assert isinstance(champ_mastery.summoner, cassiopeia.Summoner)
    assert isinstance(champ_mastery.champion, cassiopeia.Champion)

    assert champ_mastery.summoner.name == SUMMONER_NAME
    assert champ_mastery.champion.name == CHAMP_NAME


def test_mastery_returns_correct_data():
    summoner = cassiopeia.get_summoner(name=SUMMONER_NAME)
    masteries = cassiopeia.get_champion_masteries(summoner.id)
    on_champ = masteries[0]

    from_single_call = cassiopeia.get_champion_mastery(SUMMONER_NAME, on_champ.champion)

    #assert on_champ._data[cassiopeia.core.championmastery.ChampionMasteryData]._dto == from_single_call._data[cassiopeia.core.championmastery.ChampionMasteryData]._dto
    assert on_champ.points == from_single_call.points




from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError
from .common import RiotAPIService, APINotFoundError
from ...data import Platform
from ...dto.champion import ChampionDto, ChampionListDto

T = TypeVar("T")


class ChampionAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_champion_status_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(ChampionDto)
    def get_champion_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionDto:
        ChampionAPI._validate_get_champion_status_query(query, context)

        url = "https://{platform}.api.riotgames.com/lol/platform/v3/champions/{id}".format(platform=query["platform"].value.lower(), id=query["id"])
        try:
            data = self._get(url, {}, self._get_rate_limiter(query["platform"], "champions/id"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return ChampionDto(**data)

    _validate_get_many_champion_status_query = Query. \
        has("ids").as_(Iterable).also. \
        has("platform").as_(Platform)

    @get_many.register(ChampionDto)
    def get_many_champion_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionDto, None, None]:
        ChampionAPI._validate_get_many_champion_status_query(query, context)

        params = {
            "freeToPlay": False
        }

        url = "https://{platform}.api.riotgames.com/lol/platform/v3/champions".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "champions"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        statuses = {
            status["id"]: status for status in data["champions"]
        }

        def generator():
            for id in query["ids"]:
                try:
                    status = statuses[id]
                except KeyError as error:
                    raise NotFoundError("No champion exists with id \"{id}\"".format(id=id)) from error

                status["region"] = query["platform"].region.value
                yield ChampionDto(**status)

        return generator()

    _validate_get_champion_status_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("freeToPlay").with_default(False)

    @get.register(ChampionListDto)
    def get_champion_status_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionListDto:
        ChampionAPI._validate_get_champion_status_list_query(query, context)

        params = {
            "freeToPlay": query["freeToPlay"]
        }

        url = "https://{platform}.api.riotgames.com/lol/platform/v3/champions".format(platform=query["platform"].value.lower())
        try:
            data = self._get(url, params, self._get_rate_limiter(query["platform"], "champions"))
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        data["freeToPlay"] = query["freeToPlay"]
        return ChampionListDto(data)

    _validate_get_many_champion_status_list_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("freeToPlay").with_default(False)

    @get_many.register(ChampionListDto)
    def get_many_champion_status_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ChampionListDto, None, None]:
        ChampionAPI._validate_get_many_champion_status_list_query(query, context)

        params = {
            "freeToPlay": query["freeToPlay"]
        }

        def generator():
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                url = "https://{platform}.api.riotgames.com/lol/platform/v3/champions".format(platform=platform.value.lower())
                try:
                    data = self._get(url, params, self._get_rate_limiter(platform, "champions"))
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                data["freeToPlay"] = query["freeToPlay"]
                yield ChampionListDto(data)

        return generator()
